/* Cache.c b1.1   Caching control routines */

/*
	This is part of ODS2 written by Paul Nankervis,
	email address:  Paulnank@au1.ibm.com

	ODS2 is distributed freely for all members of the
	VMS community to use. However all derived works
	must maintain comments in their source to acknowledge
	the contibution of the original author.
*/

/*  Caching seems to have a big impact on performance!!!
    This version based on number of objects - probably
    should change it to use space occupied by objects.
    Currently use vanilla binary trees - could either
    use hashing or balanced trees for better performance.  */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "ssdef.h"
#include "cache.h"


#define DEBUG on
#define CACHELIM 300
#define CACHEGOAL 200

int cachecount = 0;
int cachefreecount = 0;
int cachemakecount = 0;
struct CACHE lruhead = {NULL,NULL,NULL,&lruhead,&lruhead,NULL,1};


/* cachefree - blow away item from cache - allow item to select a proxy (!)
   and adjust 'the tree' containing the item. */

struct CACHE *cachefree(struct CACHE *cacheobj,int refcount)
{
    if (cacheobj->refcount != refcount ||
	cacheobj->lstlru == NULL || cacheobj->nxtlru == NULL) {
	printf("CACHEFREE, object to delete still inuse?\n");
	return NULL;
    } else {
	register struct CACHE *proxyobj;
	while (cacheobj->getproxy != NULL) {
	    proxyobj = (*cacheobj->getproxy) (cacheobj);
	    if (proxyobj == NULL) return NULL;
	    if (proxyobj == cacheobj) break;
	    cacheobj = proxyobj;
	};
	/*        printf("CACHEFREE %x\n",cacheobj); */
	if (--cachecount < 0) printf("CACHEFREE, cache count too small\n");
	if (--cachefreecount < 0) printf("CACHEFREE, cache free count too small\n");
	cacheobj->lstlru->nxtlru = cacheobj->nxtlru;
	cacheobj->nxtlru->lstlru = cacheobj->lstlru;
	if (cacheobj->parent != NULL) {	/* Check if in tree... */
	    if (cacheobj->left == NULL) {
		if (cacheobj->right == NULL) {
		    *(cacheobj->parent) = NULL;
		} else {
		    cacheobj->right->parent = cacheobj->parent;
		    *(cacheobj->parent) = cacheobj->right;
		}
	    } else {
		if (cacheobj->right == NULL) {
		    cacheobj->left->parent = cacheobj->parent;
		    *(cacheobj->parent) = cacheobj->left;
		} else {
		    register struct CACHE *leftpath = cacheobj->left;
		    register struct CACHE *rightpath = cacheobj->right;
		    while (1) {
			if (leftpath->right == NULL) {
			    leftpath->right = cacheobj->right;
			    cacheobj->right->parent = &leftpath->right;
			    cacheobj->left->parent = cacheobj->parent;
			    *(cacheobj->parent) = cacheobj->left;
			    break;
			} else {
			    if (rightpath->left == NULL) {
				rightpath->left = cacheobj->left;
				cacheobj->left->parent = &rightpath->left;
				cacheobj->right->parent = cacheobj->parent;
				*(cacheobj->parent) = cacheobj->right;
				break;
			    } else {
				leftpath = leftpath->right;
				rightpath = rightpath->left;
			    }
			}
		    }
		}
	    }
	}
#ifdef DEBUG
	cacheobj->parent = NULL;
	cacheobj->left = NULL;
	cacheobj->right = NULL;
	cacheobj->nxtlru = NULL;
	cacheobj->lstlru = NULL;
	cacheobj->getproxy = NULL;
	cacheobj->refcount = 0;
#endif
	free(cacheobj);
	return cacheobj;
    }
}


/* cachepurge - trim size of free list */

void cachepurge()
{
    register struct CACHE *cacheobj = lruhead.lstlru;
    if (cachefreecount > CACHELIM) {
	while (cachefreecount > CACHEGOAL && cacheobj != &lruhead) {
	    register struct CACHE *lastobj = cacheobj->lstlru;
#ifdef DEBUG
	    if (cacheobj->lstlru->nxtlru != cacheobj ||
		cacheobj->nxtlru->lstlru != cacheobj ||
		*(cacheobj->parent) != cacheobj) {
		printf("CACHE pointers in bad shape! %x %x %x - %x\n",
		       cacheobj->lstlru->nxtlru,cacheobj,cacheobj->nxtlru->lstlru,*(cacheobj->parent));
	    }
#endif
	    if (cacheobj->refcount == 0) {
		if (cachefree(cacheobj,0) != lastobj) cacheobj = lastobj;
	    } else {
		printf("Object on free list with non-zero ref?\n");
	    }
	}
    }
}


/* cachemake - make a new object  */

struct CACHE *cachemake(struct CACHE **parent,unsigned length)
{
    register struct CACHE *cacheobj;
    if (length > sizeof(struct CACHE)) {
	/*        printf("CACHEMAKE %d at %x\n",length,parent); */
	cacheobj = (struct CACHE *) malloc(length);
	if (cacheobj != NULL) {
	    cacheobj->parent = parent;
	    cacheobj->left = NULL;
	    cacheobj->right = NULL;
	    cacheobj->getproxy = NULL;
	    cacheobj->refcount = 1;
	    cacheobj->lstlru = NULL;
	    cacheobj->nxtlru = NULL;
	    *parent = cacheobj;
	    cachecount++;
	    cachemakecount++;
	    if (cachefreecount > CACHELIM) cachepurge();
	}
    } else {
	cacheobj = NULL;
    }
    return cacheobj;
}


/* cacheuntouch - decrement reference count - put object on free list if zero */

unsigned cacheuntouch(struct CACHE *cacheobj,int reuse)
{
    if (cacheobj->refcount < 1) {
	printf("CACHE untouch TOO FAR!\n");
	return 64;
    } else {
	if (cacheobj->refcount-- == 1) {
#ifdef DEBUG
	    if (cacheobj->lstlru != NULL || cacheobj->lstlru != NULL) {
		printf("Cache untouch lru pointers up the creek...\n");
		return 64;
	    }
#endif
	    if (reuse) {
		cacheobj->lstlru = &lruhead;
		cacheobj->nxtlru = lruhead.nxtlru;
		lruhead.nxtlru->lstlru = cacheobj;
		lruhead.nxtlru = cacheobj;
	    } else {
		cacheobj->nxtlru = &lruhead;
		cacheobj->lstlru = lruhead.lstlru;
		lruhead.lstlru->nxtlru = cacheobj;
		lruhead.lstlru = cacheobj;
	    }
	    cachefreecount++;
	}
    }
    return 1;
}


/* cachetouch - add one to reference count - if on free list remove it! */

void cachetouch(struct CACHE *cacheobj)
{
    if (cacheobj->refcount++ == 0) {
#ifdef DEBUG
	if (cacheobj->lstlru->nxtlru != cacheobj ||
	    cacheobj->nxtlru->lstlru != cacheobj ||
	    *(cacheobj->parent) != cacheobj) {
	    printf("CACHE touch pointers in bad shape!\n");
	}
#endif
	cacheobj->lstlru->nxtlru = cacheobj->nxtlru;
	cacheobj->nxtlru->lstlru = cacheobj->lstlru;
	cacheobj->lstlru = NULL;
	cacheobj->nxtlru = NULL;
	cachefreecount--;
#ifdef DEBUG
    } else {
	if (cacheobj->lstlru != NULL || cacheobj->lstlru != NULL) {
	    printf("Cache touch lru pointers up the creek...\n");
	}
#endif
    }
}

/* cachesearch - to find or create cache entries... */

void *cachesearch(void **root,unsigned keylen,void *key,
		  int (*compare) (unsigned keylen,void *key,void *node),
		  unsigned *createsize)
{
    register struct CACHE *node,**parent = (struct CACHE **) root;
    while ((node = *parent) != NULL) {
	register cmp = (*compare) (keylen,key,(void *) node);
	if (cmp == 0) {
	    cachetouch(node);
	    return node;
	} else {
	    if (cmp < 0) {
		parent = &node->left;
	    } else {
		parent = &node->right;
	    }
	}
    }
    if (*createsize > sizeof(struct CACHE)) {
	node = cachemake(parent,*createsize);
	if (node != NULL) *createsize = 0;
    }
    return node;
}
